#include <stdlib.h>
#include <math.h>
#include "fractroam.h"

#define BOUND 35            //if we're close enough to where we're going
#define THRESH 200          //what was thresh again?
#define POINT_BUF_LENGTH 20 //max number of points to queue while roaming
#define MAXCOUNT 2000000    //max number of times to try to find interestingness

float INITIAL_XMAX=5.0;
float INITIAL_XMIN=-5.0;
float INITIAL_YMAX=4.0;
float INITIAL_YMIN=-4.0;

float MOVE_FRACTION_X = 0.05;
float MOVE_FRACTION_Y = 0.05;

//float SCREEN_FRACTION_X = 0.1;
//float SCREEN_FRACTION_Y = 0.1;

int count(int,int);
int iscomplex(int);

FractBounds queue[POINT_BUF_LENGTH];
int queue_depth[POINT_BUF_LENGTH];

FractDimension size, new_size;

int points;

short *fractal;

float xpixel, ypixel;

int pilot=1, normal=1;

FractBounds previous_bounds, bounds, target_bounds, new_bounds;

int toggle_pilot=0, toggle_style=0;

void fract_init() {
//  srandom(time(NULL)); // GNOME
  srand(time(NULL)); // WINDOWS
  if (new_bounds.xmin == new_bounds.xmax) { // hasn't been set 0==0
    bounds.xmin = INITIAL_XMIN; bounds.xmax = INITIAL_XMAX;
    bounds.ymin = INITIAL_YMIN; bounds.ymax = INITIAL_YMAX;
  } else {
    bounds.xmin = new_bounds.xmin;bounds.xmax = new_bounds.xmax;
    bounds.ymin = new_bounds.ymin;bounds.ymax = new_bounds.ymax;
    new_bounds.xmin = 0; new_bounds.xmax = 0; new_bounds.ymin = 0; new_bounds.ymax = 0;
  }
  if (new_size.width == 0) { //hasn't been set
    size.width = DEFAULT_FRACT_WIDTH; size.height = DEFAULT_FRACT_HEIGHT;
  } else {
    size.width=new_size.width; size.height=new_size.height;
    new_size.width=0; new_size.height=0;
  }
  fractal = (short *) malloc(sizeof(short)*size.height*size.width*FRACT_STEP);
  xpixel = (bounds.xmax-bounds.xmin)/size.width;
  ypixel = (bounds.ymax-bounds.ymin)/size.height;
  target_bounds.xmax = bounds.xmax;
  target_bounds.xmin = bounds.xmin;
  target_bounds.ymax = bounds.ymax;
  target_bounds.ymin = bounds.ymin;
  previous_bounds.xmax=bounds.xmax*2;
  previous_bounds.ymax=bounds.ymax*2;
  previous_bounds.xmin=bounds.xmin*2;
  previous_bounds.ymin=bounds.ymin*2;
  normal=1;
  pilot=1;
}

short* fract_get(void) {
  return fractal;
}

void fract_plot(void) {
  int x, y;
  short *pos;
  int temp;
  pos = fractal;
  if (toggle_style) {
    toggle_style=!toggle_style;
    normal=!normal;
  }
  
  for (y = size.height-1; y >=0; y--) {
    for (x = 0; x < size.width; x++) {
      temp=count(x,y)*4;
      *pos++ = temp;
      *pos++ = 256-temp;
      *pos++ = 256 % temp;
    }
  }
}

void fract_update() {
  float xbound, ybound;
  static float lastxpixel, lastypixel;
  static int outercount=0;
  float point1x=0, point1y=0, point2x=0, point2y=0;
  float width,height;
  int x, y, point=0, mycount=0;
  int forcePop=0;
  int forceNoPop=0;
  if (toggle_pilot) {
    pilot=!pilot;
    toggle_pilot=!toggle_pilot;
  }
  width=bounds.xmax-bounds.xmin;
  height=bounds.ymax-bounds.ymin;
  xpixel = width/size.width;
  ypixel = height/size.height;
  xbound = fabs(xpixel) * BOUND;
  ybound = fabs(ypixel) * BOUND;

    if ( lastxpixel == xpixel || lastypixel == ypixel
    //case 1 - a pixel is 0.0f wide, we're not going to get anywhere. fast.
         || (target_bounds.xmax == target_bounds.xmin || target_bounds.ymax == target_bounds.ymin)
         //case 2 - our target is too small
         || previous_bounds.xmin == bounds.xmin || previous_bounds.xmax == bounds.xmax || previous_bounds.ymin == bounds.ymin || previous_bounds.ymax == bounds.ymax 
          // case 3 -- we don't seem to be getting anywhere
       ) {
      forcePop=1;
    }
    if (points > 0 && (queue[points-1].xmax == 0 && queue[points-1].xmin == 0)) {
      forceNoPop=1;
    }
    //if we're within BOUND pixels for xmax, xmin, ymax, ymin -- pick a point
    if (forcePop || (fabs(bounds.xmax-target_bounds.xmax) < xbound && fabs(bounds.xmin-target_bounds.xmin) < xbound && fabs(bounds.ymax-target_bounds.ymax) < ybound && fabs(bounds.ymin-target_bounds.ymin) < ybound)) {
      //if we gotta, we gotta
      if (forcePop) {
        points=0;
        target_bounds.xmax = INITIAL_XMAX;
        target_bounds.xmin = INITIAL_XMIN;
        target_bounds.ymax = INITIAL_YMAX;
        target_bounds.ymin = INITIAL_YMIN;
      //if there is anything on the queue, 50/50 to pick off it
  //    } else if ((points > 0 && ((double)random()/RAND_MAX) > .5)){ // GNOME
     } else if ((points > 0 && ((double)rand()/RAND_MAX) > .5)){ // WINDOWS
        points--;
        target_bounds.xmax = queue[points].xmax;
        target_bounds.xmin = queue[points].xmin;
        target_bounds.ymax = queue[points].ymax;
        target_bounds.ymin = queue[points].ymin;
      //take random samples until two points are above threshold
      } else {
/*
  printf("picking a new point because we're at: \n");
  printf("width: %.11e, height: %.11e\n",width,height);
  printf("pix[%.11e,%.11e,%.11e,%.11e]\n",lastxpixel,xpixel,lastypixel,ypixel);
  printf("now[%.11e,%.11e,%.11e,%.11e]\n",bounds.xmin,bounds.xmax,bounds.ymin,bounds.ymax);
  printf("tar[%.11e,%.11e,%.11e,%.11e]\n",target_bounds.xmin,target_bounds.xmax,target_bounds.ymin,target_bounds.ymax);
*/
        while (point2y == 0 && mycount < MAXCOUNT) {
          //point = (int) fabs(((double)random()/RAND_MAX) * size.width * size.height); //GNOME
          point = (int) fabs(((double)rand()/RAND_MAX) * size.width * size.height); //WINDOWS
          mycount++;
          if (iscomplex(point)) {
            x = point % size.width;
            y = point / size.width;
            if (point1y == 0) {
              point1x = bounds.xmin + x * xpixel;
              point1y = bounds.ymin + y * ypixel;
            } else {
              point2x = bounds.xmin + x * xpixel;
              point2y = bounds.ymin + y * ypixel;
            }
          }
        }
        //set target bounds to bound that point as center
        //using SCREEN_FRACTION_X/Y
        if (mycount < MAXCOUNT) {
          target_bounds.xmax = point1x + width*(SCREEN_FRACTION_X/2);
          target_bounds.xmin = point1x - width*(SCREEN_FRACTION_X/2);
          target_bounds.ymax = point1y + height*(SCREEN_FRACTION_Y/2);
          target_bounds.ymin = point1y - height*(SCREEN_FRACTION_Y/2);
          queue[points].xmax = point2x + width*(SCREEN_FRACTION_X/2);
          queue[points].xmin = point2x - width*(SCREEN_FRACTION_X/2);
          queue[points].ymax = point2y + height*(SCREEN_FRACTION_Y/2);
          queue[points].ymin = point2y - height*(SCREEN_FRACTION_Y/2);
          points++;
        } else {
          //we're not able to find anything interesting, let's start over...
          points=0;
          target_bounds.xmax = INITIAL_XMAX;
          target_bounds.xmin = INITIAL_XMIN;
          target_bounds.ymax = INITIAL_YMAX;
          target_bounds.ymin = INITIAL_YMIN;
        }
  //printf("tar[%.11e,%.11e,%.11e,%.11e]\n\n",target_bounds.xmin,target_bounds.xmax,target_bounds.ymin,target_bounds.ymax);
      }
  outercount=0;
    }
  if (pilot) {
    //move MOVE_FRACTION_X/Y
    previous_bounds.xmax = bounds.xmax; 
    previous_bounds.xmin = bounds.xmin; 
    previous_bounds.ymax = bounds.ymax;
    previous_bounds.ymin = bounds.ymin;
    bounds.xmax += (target_bounds.xmax - bounds.xmax)* MOVE_FRACTION_X;
    bounds.xmin += (target_bounds.xmin - bounds.xmin)* MOVE_FRACTION_X;
    bounds.ymax += (target_bounds.ymax - bounds.ymax)* MOVE_FRACTION_Y;
    bounds.ymin += (target_bounds.ymin - bounds.ymin)* MOVE_FRACTION_Y;
    lastxpixel=xpixel;
    lastypixel=ypixel;
    outercount++;
    if (outercount > 200) {
  printf("stuck!?\n");    
  printf("width: %.11e, height: %.11e\n",width,height);
  printf("pix[%.11e,%.11e,%.11e,%.11e]\n",lastxpixel,xpixel,lastypixel,ypixel);
  printf("now[%.11e,%.11e,%.11e,%.11e]\n\n",bounds.xmin,bounds.xmax,bounds.ymin,bounds.ymax);
    }
  }
}

// method to see how long it takes a given point to escape reality or fail to
int count(int x, int y) {
  float z=0, zi=0, z_2=0, zi_2=0, zprev, c, ci;
  int n;
  c = bounds.xmin+x*xpixel;
  ci = bounds.ymin+y*ypixel;
  if (normal) { // normalbrot, "proper"
    for (n=0;n<256 && (z_2 + zi_2) < 4;n++) {
      zprev=z;
      z_2=z*z;
      zi_2=zi*zi;
      z = z_2 - zi_2 + c;
      zi= 2 * zprev * zi + ci;
    }
  } else { // not the "proper" but NIFTY artifacts
    for (n=0;n<256 && z < 2;n++) {
      zprev=z;
      z_2=z*z;
      zi_2=zi*zi;
      z = z_2 - zi_2 + c;
      zi= 2 * zprev * zi + ci;
    }
  }
  return n;
}

//method to see if some given point is interesting enough to zoom in on
int iscomplex(int point) {
  int x = point % size.width;
  int y = point / size.width;
  int up,left,right,down,upleft,upright,downleft,downright,center;
  center = fractal[(x+(size.width*y))*FRACT_STEP];
  if (center > THRESH) {
    up = fractal[(x+size.width*(y+1))*FRACT_STEP];
    down = fractal[(x+size.width*(y-1))*FRACT_STEP];
    left = fractal[(x-1+size.width*y)*FRACT_STEP];
    right = fractal[(x+1+size.width*y)*FRACT_STEP];
    upleft = fractal[(x-1+size.width*(y+1))*FRACT_STEP];
    downleft = fractal[(x-1+size.width*(y-1))*FRACT_STEP];
    upright = fractal[(x+1+size.width*(y+1))*FRACT_STEP];
    downright = fractal[(x+1+size.width*(y-1))*FRACT_STEP];
    return (center != up && center != down && center != left && center != right && center != upleft && center != downleft && center != upright && center != downright && upleft != upright && upleft != downleft && upright != downright);
  }
  return 0;
}

FractBounds fract_bounds_get(void) {
  return bounds;
}

void fract_bounds_set(FractBounds fb) {
  if (!pilot) {
    //new_bounds.xmin = fb.xmin; new_bounds.xmax = fb.xmax;
    //new_bounds.ymin = fb.ymin; new_bounds.ymax = fb.ymax;
    bounds.xmin=fb.xmin; bounds.xmax=fb.xmax;
    bounds.ymin=fb.ymin; bounds.ymax=fb.ymax;
  }
}

void fract_size_set(FractDimension fd) {
  new_size.width = fd.width; new_size.height = fd.height;
}

FractDimension fract_size_get(void) {
  return size;
}

void fract_pilot_toggle(void) {
  toggle_pilot=!toggle_pilot;
}

void fract_style_toggle(void) {
  toggle_style=!toggle_style;
}

void fract_interrupt(void) {
}

void fract_center(double x, double y) {
  float width, height, perx, pery, xcenter, ycenter;
  if (!pilot) {
    width=bounds.xmax-bounds.xmin; height=bounds.ymax-bounds.ymin;
    perx=x/size.width; pery=y/size.height;
    xcenter=(bounds.xmax+bounds.xmin)/2; ycenter=(bounds.ymax+bounds.ymin)/2;
    //new center is the portion of the width, shifted
    xcenter+=(perx-.5)*width; ycenter+=(pery-.5)*height;
    bounds.xmax=xcenter+width/2; bounds.xmin=xcenter-width/2;
    bounds.ymax=ycenter+height/2; bounds.ymin=ycenter-height/2;
  }
}

void fract_move(int up, int left, int in) {
  float width, height;
  if (!pilot) {
    width  = bounds.xmax - bounds.xmin;
    height = bounds.ymax - bounds.ymin;
    //move up
    bounds.ymax += (up*height)  * SCREEN_FRACTION_Y;
    bounds.ymin += (up*height)  * SCREEN_FRACTION_Y;
    //move left
    bounds.xmax += (left*width) * SCREEN_FRACTION_X;
    bounds.xmin += (left*width) * SCREEN_FRACTION_X;
    //move "in"
    bounds.xmax += (width*in)   * SCREEN_FRACTION_X;
    bounds.xmin -= (width*in)   * SCREEN_FRACTION_X;
    bounds.ymax += (height*in)  * SCREEN_FRACTION_Y;
    bounds.ymin -= (height*in)  * SCREEN_FRACTION_Y;
		previous_bounds.xmax = bounds.xmax;
		previous_bounds.xmin = bounds.xmin;
		previous_bounds.ymax = bounds.ymax;
		previous_bounds.ymin = bounds.ymin;
  }
}
